#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
.. codeauthor:: Cédric Dumay <cedric.dumay@gmail.com>


"""
import logging
from kser.entry import Entrypoint
from kser.schemas import Message

logger = logging.getLogger(__name__)


class Task(Entrypoint):
    """"""

    @classmethod
    def init_by_id(cls, _id):
        """Load task by its ID"""

    def __init__(self, uuid=None, params=None, status="PENDING", result=None,
                 metadata=None):
        """ A task is a :class:`kser.entry.Entrypoint` with additional attributes.

        :param str uuid: task unique identifier
        :param dict params: task parameter
        :param str status: task status
        :param cdumay_result.Result result: forwarded result from a previous task
        :param dict metadata: task context
        """
        self.status = status
        Entrypoint.__init__(
            self, uuid=uuid, params=params, result=result, metadata=metadata
        )

    def get_attr(self, item):
        attr = "{}Id".format(item)
        return attr, getattr(getattr(self, item, self), attr, None)

    def log(self, message, level=logging.INFO, *args, **kwargs):
        """ Send log entry

        :param str message: log message
        :param int level: `Logging level <https://docs.python.org/3/library/logging.html#levels>`_
        :param list args: log record arguments
        :param dict kwargs: log record key argument
        """
        msg = "{}.{}: {}[{}]: {}".format(
            self.__class__.__name__, self.status, self.__class__.path,
            self.uuid, message
        )
        extra = kwargs.pop("extra", dict())
        extra.update(dict(kmsg=Message(
            self.uuid, entrypoint=self.__class__.path, params=self.params,
            metadata=self.metadata
        ).dump()))

        return logger.log(
            level=level, msg=msg, extra=extra, *args, **kwargs
        )

    def __repr__(self):
        return "Task [{}](uuid={},status={}, params={})".format(
            self.__class__.path, self.uuid, self.status, self.params
        )



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
.. codeauthor:: Cédric Dumay <cedric.dumay@gmail.com>


"""
import logging
import os
import time

from kafka import KafkaConsumer
from kser import KSER_METRICS_ENABLED
from kser.controller import Controller

logger = logging.getLogger(__name__)


class Consumer(object):
    REGISTRY = Controller

    def __init__(self, config, topics):
        self.client = KafkaConsumer(**config)
        self.client.subscribe(topics)
        self.clean_lock()

    def __del__(self):
        for lockfile in (os.environ['LOCK_FILE'], os.environ['RUNNING_FILE']):
            if os.path.exists(lockfile):
                logger.debug("Cleaning existing lock file: {}".format(lockfile))
                os.remove(lockfile)

    def clean_lock(self):
        if os.path.exists(os.environ['LOCK_FILE']):
            logger.debug("Cleaning existing pause file: {}".format(
                os.environ['LOCK_FILE']
            ))
            os.remove(os.environ['LOCK_FILE'])

    def is_active(self):
        return not os.path.exists(os.environ['LOCK_FILE'])

    def run(self):
        """ Run consumer
        """
        if KSER_METRICS_ENABLED == "yes":
            from prometheus_client import start_http_server
            logger.info("Metric.Starting...")
            start_http_server(
                os.getenv("KSER_METRICS_PORT", 8888),
                os.getenv("KSER_METRICS_ADDRESS", "0.0.0.0")
            )

        logger.info("{}.Starting...".format(self.__class__.__name__))

        while True:
            if self.is_active() is True:
                msg = next(self.client)
                data = msg.value.decode('utf-8')
                if self.client.config['enable_auto_commit'] is False:
                    self.client.commit()
                    logger.debug("{}: Manual commit done.".format(
                        self.__class__.__name__
                    ))
                self.REGISTRY.run(data)
            else:
                logger.warning("Consumer is paused")
                time.sleep(60)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
.. codeauthor:: Cédric Dumay <cedric.dumay@gmail.com>


"""
from kafka import KafkaProducer
from cdumay_result import Result
from kser.controller import BaseController


class Producer(BaseController):
    def __init__(self, config):
        """ Create new Producer instance using provided configuration dict.

        :param dict config: configuration
        """
        self.client = KafkaProducer(**config)

    def bulk_send(self, topic, kmsgs, timeout=60):
        """ Send a batch of messages

        :param str topic: a kafka topic
        :param ksr.transport.Message kmsgs: Messages to serialize
        :param int timeout: Timeout in seconds
        :return: Execution result
        :rtype: kser.result.Result
        """

        try:
            for kmsg in kmsgs:
                self.client.send(
                    topic, self._onmessage(kmsg).dumps().encode("UTF-8")
                )
            self.client.flush(timeout=timeout)
            return Result(stdout="{} message(s) sent".format(len(kmsgs)))

        except Exception as exc:
            return Result.from_exception(exc)

    # noinspection PyUnusedLocal
    def send(self, topic, kmsg, timeout=60):
        """ Send the message into the given topic

        :param str topic: a kafka topic
        :param ksr.transport.Message kmsg: Message to serialize
        :param int timeout: Timeout in seconds (not used in proto producer)
        :return: Execution result
        :rtype: kser.result.Result
        """

        result = Result(uuid=kmsg.uuid)
        try:
            self.client.send(
                topic, self._onmessage(kmsg).dumps().encode("UTF-8")
            )
            result.stdout = "Message {}[{}] sent in {}".format(
                kmsg.entrypoint, kmsg.uuid, topic
            )
            self.client.flush()

        except Exception as exc:
            result = Result.from_exception(exc, kmsg.uuid)

        finally:
            if result.retcode < 300:
                return self._onsuccess(kmsg=kmsg, result=result)
            else:
                return self._onerror(kmsg=kmsg, result=result)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
.. codeauthor:: Cédric Dumay <cedric.dumay@gmail.com>


"""
import logging
import opentracing

from kser.sequencing.operation import Operation
from kser.tracing.proxy import KserSpan

logger = logging.getLogger(__name__)


class OpentracingOperation(Operation):
    def unsafe_execute(self, result=None):
        with opentracing.tracer.start_span(
                obj=self, child_of=KserSpan.extract_span(self),
                span_factory=KserSpan) as span:
            for task in self.tasks:
                KserSpan.inject_span(span, task)
                result = task.unsafe_execute(result=result)
                if result.retcode != 0:
                    return self._onerror(result)

            self.result = self._onsuccess(self._postrun(result=result))
            span.obj = self
            return self.result



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
.. codeauthor:: Cédric Dumay <cedric.dumay@gmail.com>


"""
from cdumay_error import ValidationError
import marshmallow.exceptions
from marshmallow import Schema, fields, EXCLUDE
from cdumay_result import ResultSchema, Result


class BaseSchema(Schema):
    class Meta:
        unknown = EXCLUDE

    uuid = fields.String(required=True)
    entrypoint = fields.String(required=True)
    params = fields.Dict(default=dict)


class Base(object):
    MARSHMALLOW_SCHEMA = BaseSchema()

    def __init__(self, uuid, entrypoint, params=None):
        self.uuid = uuid
        self.entrypoint = entrypoint
        self.params = params if params else dict()

    def dump(self):
        """description of dump"""
        return self.MARSHMALLOW_SCHEMA.dump(self)

    def dumps(self):
        """description of dumps"""
        return self.MARSHMALLOW_SCHEMA.dumps(self)

    def __str__(self):
        return str(self.dump())


class MessageSchema(BaseSchema):
    result = fields.Nested(ResultSchema, missing=None)
    metadata = fields.Dict()


class Message(Base):
    MARSHMALLOW_SCHEMA = MessageSchema()

    @classmethod
    def loads(cls, json_data):
        """description of load"""
        try:
            return cls(**cls.MARSHMALLOW_SCHEMA.loads(json_data))
        except marshmallow.exceptions.ValidationError as exc:
            raise ValidationError("Failed to load message", extra=exc.args[0])

    def __init__(self, uuid, entrypoint, params=None, result=None,
                 metadata=None):
        Base.__init__(self, uuid, entrypoint, params)
        if result:
            if isinstance(result, Result):
                self.result = result
            else:
                self.result = Result(**result)
        else:
            self.result = Result(uuid=uuid)
        self.metadata = metadata or dict()

    def __repr__(self):
        """"""
        return "Message<uuid='{}', entrypoint='{}', result={}>".format(
            self.uuid, self.entrypoint, self.result
        )



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
.. codeauthor:: Cédric Dumay <cedric.dumay@gmail.com>


"""
import logging

from cdumay_error import ValidationError
from cdumay_result import Result, ResultSchema
from kser.schemas import Message
from kser.sequencing.task import Task

logger = logging.getLogger(__name__)


class Operation(Task):
    TASKS = ()

    def check_required_params(self):
        """ Check if all required parameters are set"""
        for param in self.REQUIRED_FIELDS:
            if param not in self.params:
                raise ValidationError("Missing parameter: {} for {}".format(
                    param, self.__class__.path
                ))

        for child in self.TASKS:
            for param in child.REQUIRED_FIELDS:
                if param not in self.params:
                    raise ValidationError(
                        "Missing parameter: {} for {}".format(param, child.path)
                    )

    @classmethod
    def new(cls, **kwargs):
        return cls(**cls.parse_inputs(**kwargs))

    @classmethod
    def parse_inputs(cls, **kwargs):
        return kwargs

    def __init__(self, uuid=None, status="PENDING", params=None, tasks=None,
                 result=None, metadata=None):
        Task.__init__(self, uuid=uuid, params=params, status=status,
                      result=result, metadata=metadata)
        self.tasks = tasks or list()

    def __repr__(self):
        return "Operation [{}](id={},status={})".format(
            self.__class__.__name__, self.uuid, self.status
        )

    def _set_status(self, status, result=None):
        """ update operation status

        :param str status: New status
        :param cdumay_result.Result result: Execution result
        """
        logger.info(
            "{}.SetStatus: {}[{}] status update '{}' -> '{}'".format(
                self.__class__.__name__, self.__class__.path, self.uuid,
                self.status, status
            ),
            extra=dict(
                kmsg=Message(
                    self.uuid, entrypoint=self.__class__.path,
                    params=self.params
                ).dump()
            )
        )
        return self.set_status(status, result)

    # noinspection PyUnusedLocal
    def set_status(self, status, result=None):
        """ update operation status

        :param str status: New status
        :param cdumay_result.Result result: Execution result
        """
        self.status = status

    def add_task(self, task):
        """ add task to operation

        :param kser.sequencing.task.Task task: task to add
        """
        self.tasks.append(task)

    def _prebuild(self, **kwargs):
        logger.debug(
            "{}.PreBuild: {}[{}]: {}".format(
                self.__class__.__name__, self.__class__.path, self.uuid, kwargs
            ),
            extra=dict(
                kmsg=Message(
                    self.uuid, entrypoint=self.__class__.path,
                    params=self.params
                ).dump()
            )
        )
        self.check_required_params()
        return self.prebuild(**kwargs)

    # noinspection PyMethodMayBeStatic
    def prebuild(self, **kwargs):
        """ to implement, perform check before the operation creation
        """
        return kwargs

    def _prerun(self):
        """ To execute before running message
        """
        self.check_required_params()
        self._set_status("RUNNING")
        logger.debug(
            "{}.PreRun: {}[{}]: running...".format(
                self.__class__.__name__, self.__class__.path, self.uuid
            ),
            extra=dict(
                kmsg=Message(
                    self.uuid, entrypoint=self.__class__.path,
                    params=self.params
                ).dump()
            )
        )
        return self.prerun()

    def _onsuccess(self, result):
        """ To execute on execution success
        :param cdumay_result.Result result: Execution result
        :return: Execution result
        :rtype: cdumay_result.Result
        """
        self._set_status("SUCCESS", result)
        logger.info(
            "{}.Success: {}[{}]: {}".format(
                self.__class__.__name__, self.__class__.path, self.uuid, result
            ),
            extra=dict(
                kmsg=Message(
                    self.uuid, entrypoint=self.__class__.path,
                    params=self.params
                ).dump(),
                kresult=ResultSchema().dump(result) if result else dict()
            )
        )
        return self.onsuccess(result)

    def _onerror(self, result):
        """ To execute on execution failure

        :param cdumay_result.Result result: Execution result
        :return: Execution result
        :rtype: cdumay_result.Result
        """
        self._set_status("FAILED", result)
        extra = dict(
            kmsg=Message(
                self.uuid, entrypoint=self.__class__.path, params=self.params
            ).dump(),
            kresult=ResultSchema().dump(result) if result else dict()
        )
        error = result.search_value("error")
        if error:
            extra['error'] = error

        logger.error(
            "{}.Failed: {}[{}]: {}".format(
                self.__class__.__name__, self.__class__.path, self.uuid, result
            ), extra=extra
        )
        return self.onerror(result)

    def unsafe_execute(self, result=None):
        self._prerun()
        for task in self.tasks:
            if task.status != 'SUCCESS':
                result = task.unsafe_execute(result=result)
                if result.retcode != 0:
                    return self._onerror(result)
            else:
                result = task.result

        return self._onsuccess(result=result)

    def display(self):
        """ dump operation

        """
        print("{}".format(self))
        for task in self.tasks:
            print("  - {}".format(task))

    def next(self, task):
        """ Find the next task

        :param kser.sequencing.task.Task task: previous task
        :return: The next task
        :rtype: kser.sequencing.task.Task or None
        """
        uuid = str(task.uuid)
        for idx, otask in enumerate(self.tasks[:-1]):
            if otask.uuid == uuid:
                if self.tasks[idx + 1].status != 'SUCCESS':
                    return self.tasks[idx + 1]
                else:
                    uuid = self.tasks[idx + 1].uuid

    def launch_next(self, task=None, result=None):
        """ Launch next task or finish operation

        :param kser.sequencing.task.Task task: previous task
        :param cdumay_result.Result result: previous task result

        :return: Execution result
        :rtype: cdumay_result.Result
        """
        if task:
            next_task = self.next(task)
            if next_task:
                return next_task.send(result=result)
            else:
                return self.set_status(task.status, result)
        elif len(self.tasks) > 0:
            return self.tasks[0].send(result=result)
        else:
            return Result(retcode=1, stderr="Nothing to do, empty operation !")

    def launch(self):
        """ Send the first task

        :return: Execution result
        :rtype: cdumay_result.Result
        """
        return self.launch_next()

    def finalize(self):
        """To implement, post build actions (database mapping ect...)

        :return: the controller
        :rtype: kser.sequencing.operation.Operation
        """
        return self

    def _build_tasks(self, **kwargs):
        """

        :param dict kwargs: tasks parameters (~=context)
        :return: list of tasks
        :rtype: list(kser.sequencing.operation.Operation)
        """
        tasks = self.build_tasks(**kwargs)
        logger.debug(
            "{}.BuildTasks: {} task(s) found".format(
                self.__class__.__name__, len(tasks)
            ),
            extra=dict(
                kmsg=Message(
                    self.uuid, entrypoint=self.__class__.path,
                    params=self.params
                ).dump()
            )
        )
        return tasks

    # noinspection PyMethodMayBeStatic
    # noinspection PyUnusedLocal
    def build_tasks(self, **kwargs):
        """ to implement

        :param dict kwargs: tasks parameters (~=context)
        :return: list of tasks
        :rtype: list(kser.sequencing.task.Task)
        """
        return list()

    def compute_tasks(self, **kwargs):
        """ perfrom checks and build tasks

        :return: list of tasks
        :rtype: list(kser.sequencing.operation.Operation)
        """
        params = self._prebuild(**kwargs)
        if not params:
            params = dict(kwargs)

        return self._build_tasks(**params)

    def build(self, **kwargs):
        """ create the operation and associate tasks

        :param dict kwargs: operation data
        :return: the controller
        :rtype: kser.sequencing.controller.OperationController
        """
        self.tasks += self.compute_tasks(**kwargs)
        return self.finalize()

    def send(self):
        """ Send operation to Kafka

        :return: The operation
        :rtype: kser.sequencing.operation.Operation
        """



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
.. codeauthor:: Cédric Dumay <cedric.dumay@gmail.com>


"""

import logging

import opentracing
from kser.sequencing.task import Task
from kser.tracing.proxy import KserSpan

logger = logging.getLogger(__name__)


class OpentracingTask(Task):
    def unsafe_execute(self, result=None):
        """ un-wrapped execution, can raise excepetion

        :return: Execution result
        :rtype: kser.result.Result
        """
        if result:
            self.result += result

        with opentracing.tracer.start_span(
                obj=self, child_of=KserSpan.extract_span(self),
                span_factory=KserSpan) as span:
            self.result = self._onsuccess(self._postrun(self._run()))
            span.obj = self
            return self.result



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
.. codeauthor:: Cédric Dumay <cedric.dumay@gmail.com>


"""
import logging
from uuid import uuid4

from cdumay_error import ValidationError
from cdumay_result import Result, ResultSchema
from kser import KSER_TASK_COUNT, __hostname__, KSER_METRICS_ENABLED, \
    KSER_TASKS_STATUS
from kser.schemas import Message

logger = logging.getLogger(__name__)


class EntrypointMeta(type):
    @property
    def path(cls):
        return "{}.{}".format(cls.__module__, cls.__name__)


class Entrypoint(object, metaclass=EntrypointMeta):
    REQUIRED_FIELDS = []

    def __init__(self, uuid=None, params=None, result=None, metadata=None):
        self.uuid = uuid or str(uuid4())
        self.params = params or dict()
        self.result = result or Result(uuid=self.uuid)
        self.metadata = metadata or dict()
        self._post_init()

    def label(self, action=None):
        return "{}[{}]{}".format(
            self.__class__.__name__, self.uuid,
            " - {}".format(action) if action else ""
        )

    def _post_init(self):
        """A post init trigger"""
        try:
            return self.postinit()
        except Exception as exc:
            return self._onerror(Result.from_exception(exc, uuid=self.uuid))

    def postinit(self):
        """"""

    def check_required_params(self):
        """ Check if all required parameters are set"""
        for param in self.REQUIRED_FIELDS:
            if param not in self.params:
                raise ValidationError("Missing parameter: {}".format(param))

    def _onsuccess(self, result):
        """ To execute on execution success

        :param kser.result.Result result: Execution result
        :return: Execution result
        :rtype: kser.result.Result
        """
        if KSER_METRICS_ENABLED == "yes":
            KSER_TASKS_STATUS.labels(
                __hostname__, self.__class__.path, 'SUCCESS'
            ).inc()
        if result:
            result = self.result + result
        else:
            result = self.result
        logger.info(
            "{}.Success: {}[{}]: {}".format(
                self.__class__.__name__, self.__class__.path, self.uuid, result
            ),
            extra=dict(
                kmsg=Message(
                    self.uuid, entrypoint=self.__class__.path,
                    params=self.params, metadata=self.metadata
                ).dump(),
                kresult=ResultSchema().dump(result) if result else dict()
            )
        )
        return self.onsuccess(result)

    def log(self, message, level=logging.INFO, *args, **kwargs):
        msg = "{}.MESSAGE: {}[{}]: {}".format(
            self.__class__.__name__, self.__class__.path, self.uuid, message
        )
        return logger.log(level=level, msg=msg, *args, **kwargs)

    def onsuccess(self, result):
        """ To execute on execution success

        :param kser.result.Result result: Execution result
        :return: Execution result
        :rtype: kser.result.Result
        """
        return result

    def _onerror(self, result):
        """ To execute on execution failure

        :param kser.result.Result result: Execution result
        :return: Execution result
        :rtype: kser.result.Result
        """
        if KSER_METRICS_ENABLED == "yes":
            KSER_TASKS_STATUS.labels(
                __hostname__, self.__class__.path, 'FAILED'
            ).inc()

        if result:
            result = self.result + result
        else:
            result = self.result

        extra = dict(
            kmsg=Message(
                self.uuid, entrypoint=self.__class__.path,
                params=self.params, metadata=self.metadata
            ).dump(),
            kresult=ResultSchema().dump(result) if result else dict()
        )

        if result:
            error = result.search_value("error")
            if error:
                extra['error'] = error

        logger.error(
            "{}.Failed: {}[{}]: {}".format(
                self.__class__.__name__, self.__class__.path, self.uuid, result
            ), extra=extra
        )
        return self.onerror(result)

    def onerror(self, result):
        """ To implement

        :param kser.result.Result result: Execution result
        :return: Execution result
        :rtype: kser.result.Result
        """
        return result

    def _prerun(self):
        """ To execute before running message

        :return: Kafka message
        :rtype: kser.schemas.Message
        """
        logger.debug(
            "{}.PreRun: {}[{}]".format(
                self.__class__.__name__, self.__class__.path, self.uuid
            ),
            extra=dict(
                kmsg=Message(
                    self.uuid, entrypoint=self.__class__.path,
                    params=self.params, metadata=self.metadata
                ).dump()
            )
        )
        self.check_required_params()
        return self.prerun()

    def prerun(self):
        """ To implement"""

    def _postrun(self, result):
        """ To execute after exection

        :param kser.result.Result result: Execution result
        :return: Execution result
        :rtype: kser.result.Result
        """
        logger.debug(
            "{}.PostRun: {}[{}]".format(
                self.__class__.__name__, self.__class__.path, self.uuid
            ),
            extra=dict(
                kmsg=Message(
                    self.uuid, entrypoint=self.__class__.path,
                    params=self.params, metadata=self.metadata
                ).dump()
            )
        )
        return self.postrun(result)

    def postrun(self, result):
        """ To implement

        :param kser.result.Result result: Execution result
        :return: Execution result
        :rtype: kser.result.Result
        """
        return result

    def _run(self):
        """ Execution body

        :return: Execution result
        :rtype: kser.result.Result
        """
        if KSER_METRICS_ENABLED == "yes":
            KSER_TASK_COUNT.inc()

        logger.debug(
            "{}.Run: {}[{}]".format(
                self.__class__.__name__, self.__class__.path, self.uuid
            ),
            extra=dict(
                kmsg=Message(
                    self.uuid, entrypoint=self.__class__.path,
                    params=self.params, metadata=self.metadata
                ).dump()
            )
        )
        return self.run()

    def run(self):
        """ To implement

        :return: Execution result
        :rtype: kser.result.Result
        """
        raise NotImplemented("Task '{}' not implemented".format(
            self.__class__.path
        ))

    def unsafe_execute(self, result=None):
        """ un-wrapped execution, can raise excepetion

        :return: Execution result
        :rtype: kser.result.Result
        """
        if result:
            self.result += result

        self._prerun()
        return self._onsuccess(self._postrun(self._run()))

    def execute(self, result=None):
        """ Execution 'wrapper' to make sure that it return a result

        :return: Execution result
        :rtype: kser.result.Result
        """
        try:
            return self.unsafe_execute(result=result)
        except Exception as exc:
            return self._onerror(Result.from_exception(exc, uuid=self.uuid))

    # noinspection PyPep8Naming
    def to_Message(self, result=None):
        """ Entrypoint -> Message

        :param kser.result.Result result: Execution result
        :return: Kafka message
        :rtype kser.schemas.Message
        """
        return Message(
            uuid=self.uuid, entrypoint=self.__class__.path, params=self.params,
            result=result if result else self.result, metadata=self.metadata
        )

    # noinspection PyPep8Naming
    @classmethod
    def from_Message(cls, kmsg):
        """ Message -> Entrypoint

        :param kser.schemas.Message kmsg: Kafka message
        :return: a entrypoint
        :rtype kser.entry.Entrypoint
        """
        return cls(
            uuid=kmsg.uuid, params=kmsg.params, result=kmsg.result,
            metadata=kmsg.metadata
        )



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
.. codeauthor:: Cédric Dumay <cedric.dumay@gmail.com>


"""
import socket
import os

__hostname__ = socket.gethostname()

os.environ.setdefault("LOCK_PATH", "/var/run")
os.environ.setdefault("RUNNING_FILE", os.path.join(
    os.environ['LOCK_PATH'], "kser-{}-run.lock".format(__hostname__)
))
os.environ.setdefault("LOCK_FILE", os.path.join(
    os.environ['LOCK_PATH'], "kser-{}-pause.lock".format(__hostname__)
))

KSER_METRICS_ENABLED = os.getenv("KSER_METRICS_ENABLED", "no")
KSER_TASK_COUNT = None
KSER_TASKS_STATUS = None

if KSER_METRICS_ENABLED == "yes":
    from prometheus_client import Counter, Gauge

    KSER_TASK_COUNT = Gauge('kser_task_count', 'KSER tasks launched')
    KSER_TASKS_STATUS = Counter(
        'kser_task_result', 'KSER task status', ['hostname', 'name', 'status']
    )



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
.. codeauthor:: Cédric Dumay <cedric.dumay@gmail.com>


"""
import logging
import os

from cdumay_error import ValidationError, from_exc
from kser.entry import Entrypoint
from cdumay_result import Result, ResultSchema
from kser.schemas import Message

logger = logging.getLogger(__name__)

os.environ.setdefault("RUNNING_FILE", "/var/run/kser-run.lock")


class BaseController(object):
    @classmethod
    def _onsuccess(cls, kmsg, result):
        """ To execute on execution success

        :param kser.schemas.Message kmsg: Kafka message
        :param kser.result.Result result: Execution result
        :return: Execution result
        :rtype: kser.result.Result
        """
        logger.info(
            "{}.Success: {}[{}]: {}".format(
                cls.__name__, kmsg.entrypoint, kmsg.uuid, result
            ),
            extra=dict(
                kmsg=kmsg.dump(),
                kresult=ResultSchema().dump(result) if result else dict()
            )
        )
        return cls.onsuccess(kmsg, result)

    # noinspection PyUnusedLocal
    @classmethod
    def onsuccess(cls, kmsg, result):
        """ To execute on execution success

        :param kser.schemas.Message kmsg: Kafka message
        :param kser.result.Result result: Execution result
        :return: Execution result
        :rtype: kser.result.Result
        """
        return result

    @classmethod
    def _onerror(cls, kmsg, result):
        """ To execute on execution failure

        :param kser.schemas.Message kmsg: Kafka message
        :param kser.result.Result result: Execution result
        :return: Execution result
        :rtype: kser.result.Result
        """
        logger.error(
            "{}.Failed: {}[{}]: {}".format(
                cls.__name__, kmsg.entrypoint, kmsg.uuid, result
            ),
            extra=dict(
                kmsg=kmsg.dump(),
                kresult=ResultSchema().dump(result) if result else dict()
            )
        )
        return cls.onerror(kmsg, result)

    # noinspection PyUnusedLocal
    @classmethod
    def onerror(cls, kmsg, result):
        """ To implement

        :param kser.schemas.Message kmsg: Kafka message
        :param kser.result.Result result: Execution result
        :return: Execution result
        :rtype: kser.result.Result
        """
        return result

    @classmethod
    def _onmessage(cls, kmsg):
        """ Call on received message

        :param kser.schemas.Message kmsg: Kafka message
        :return: Kafka message
        :rtype: kser.schemas.Message
        """
        logger.debug(
            "{}.ReceivedMessage {}[{}]".format(
                cls.__name__, kmsg.entrypoint, kmsg.uuid
            ),
            extra=dict(kmsg=kmsg.dump())
        )
        return cls.onmessage(kmsg)

    @classmethod
    def onmessage(cls, kmsg):
        """ To implement

        :param kser.schemas.Message kmsg: Kafka message
        :return: Kafka message
        :rtype: kser.schemas.Message
        """
        return kmsg


class Controller(BaseController):
    ENTRYPOINTS = dict()
    TRANSPORT = Message

    @classmethod
    def start_processing(cls, kmsg):
        with open(os.environ['RUNNING_FILE'], 'w') as file:
            file.write("{}[{}]".format(kmsg.entrypoint, kmsg.uuid))

    @classmethod
    def stop_processing(cls):
        with open(os.environ['RUNNING_FILE'], 'w') as file:
            file.write("")

    @classmethod
    def register(cls, name, entrypoint):
        """ Register a new entrypoint

        :param str name: Key used by messages
        :param kser.entry.Entrypoint entrypoint: class to load
        :raises ValidationError: Invalid entry
        """
        if not issubclass(entrypoint, Entrypoint):
            raise ValidationError(
                "Invalid type for entry '{}', MUST implement "
                "kser.entry.Entrypoint".format(name),
                extra=dict(entrypoint=name)
            )
        cls.ENTRYPOINTS[name] = entrypoint
        logger.debug("{}.Registered: {}".format(cls.__name__, name))

    @classmethod
    def run(cls, raw_data):
        """description of run"""
        logger.debug("{}.ReceivedFromKafka: {}".format(
            cls.__name__, raw_data
        ))
        try:
            kmsg = cls._onmessage(cls.TRANSPORT.loads(raw_data))
        except Exception as exc:
            error = from_exc(exc)
            logger.error(
                "{}.ImportError: Failed to load data from kafka: {}".format(
                    cls.__name__, exc
                ),
                extra=dict(kafka_raw_data=raw_data, error=error.to_dict())
            )
            return Result.from_error(error)

        try:
            cls.start_processing(kmsg)
            if kmsg.entrypoint not in cls.ENTRYPOINTS:
                raise ValidationError(
                    "Entrypoint '{}' not registred".format(kmsg.entrypoint),
                    extra=dict(
                        uuid=kmsg.uuid, entrypoint=kmsg.entrypoint,
                        allowed=list(cls.ENTRYPOINTS.keys())
                    )
                )

            result = cls.ENTRYPOINTS[kmsg.entrypoint].from_Message(
                kmsg
            ).execute()

        except Exception as exc:
            result = Result.from_exception(exc, kmsg.uuid)

        finally:
            cls.stop_processing()
            # noinspection PyUnboundLocalVariable
            if result and result.retcode < 300:
                return cls._onsuccess(kmsg=kmsg, result=result)
            else:
                return cls._onerror(kmsg=kmsg, result=result)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
.. codeauthor:: Cédric Dumay <cedric.dumay@gmail.com>


"""
import json
import re

import opentracing
from cdumay_opentracing import Span
from cdumay_result import ResultSchema


class KserSpan(Span):
    EXCLUDE_PATTERNS = ['.*(password|PASSWORD).*', '.*(token|TOKEN).*']

    @classmethod
    def name(cls, obj):
        return str(obj.__class__.path)

    @classmethod
    def extract_span(cls, obj):
        trace = obj.metadata.get('__parent-span__', dict())
        if len(trace) > 0:
            return cls.span_from_dict(trace)

    @classmethod
    def inject_span(cls, span, obj):
        obj.metadata['__parent-span__'] = cls.span_serialize(span)

    @classmethod
    def extract(cls, obj):
        """ Extract span context from the given object

        :param Any obj: Object to use as context
        :return: a SpanContext instance extracted from the inner span object or None if no
            such span context could be found.
        """
        span = cls.extract_span(obj)
        if span:
            return span.context

    @classmethod
    def inject(cls, span, obj):
        """ Injects the span context into a `carrier` object.

        :param opentracing.span.SpanContext span: the SpanContext instance
        :param Any obj: Object to use as context
        """
        obj.metadata['__parent-span__'] = dict()
        cls.inject_span(span, obj.metadata['__parent-span__'])

    @classmethod
    def extract_tags(cls, obj):
        """ Extract tags from the given object

        :param Any obj: Object to use as context
        :return: Tags to add on span
        :rtype: dict
        """
        return dict(uuid=obj.uuid, entrypoint=obj.__class__.path)

    @classmethod
    def _postrun(cls, span, obj, **kwargs):
        """ Trigger to execute just before closing the span

        :param opentracing.span.Span  span: the SpanContext instance
        :param Any obj: Object to use as context
        :param dict kwargs: additional data
        """
        for key, value in ResultSchema().dump(obj.result).items():
            if isinstance(value, dict):
                try:
                    flat_data = cls.filter_keys(
                        cls.fix_additional_fields(value)
                    )
                    span.set_tag("result.{}".format(key), json.dumps(flat_data))
                except Exception:
                    span.set_tag("result.{}".format(key), "N/A")
            elif isinstance(value, (list, tuple)):
                try:
                    span.set_tag("result.{}".format(key), json.dumps(value))
                except Exception:
                    try:
                        span.set_tag("result.{}".format(key), value)
                    except Exception:
                        span.set_tag("result.{}".format(key), "N/A")
            else:
                span.set_tag("result.{}".format(key), value)

    @staticmethod
    def fix_additional_fields(data):
        """description of fix_additional_fields"""
        result = dict()
        for key, value in data.items():
            if isinstance(value, dict):
                result.update(KserSpan.to_flat_dict(key, value))
            else:
                result[key] = value
        return result

    @staticmethod
    def key_path(*args):
        """description of key_path"""
        return "_".join(args)

    @staticmethod
    def to_flat_dict(prefix, data):
        flat_result = dict()
        for dkey, dvalue in data.items():
            path = KserSpan.key_path(prefix, dkey)
            if isinstance(dvalue, dict):
                flat_result.update(KserSpan.to_flat_dict(path, dvalue))
            else:
                flat_result[path] = dvalue
        return flat_result

    @classmethod
    def filter_keys(cls, data):
        """Filter GELF record keys using exclude_patterns

        :param dict data: Log record has dict
        :return: the filtered log record
        :rtype: dict
        """
        keys = list(data.keys())
        for pattern in cls.EXCLUDE_PATTERNS:
            for key in keys:
                if re.match(pattern, key):
                    keys.remove(key)
        return dict(filter(lambda x: x[0] in keys, data.items()))
